'''
Created on 2022-03-24

@author: wf
'''

class Version(object):
    '''
    Version handling for py-3rdparty-mediawiki
    '''
    name="py-3rdparty-mediawiki"
    version='0.7.2'
    date = '2020-10-31'
    updated = '2022-11-24'


'''
Created on 2020-11-01

@author: wf
'''

import os
import sys
from os.path import isdir
from os import makedirs
from wikibot.crypt import Crypt
import datetime
from wikibot.version import Version

from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
from pathlib import Path
import getpass

class WikiUser(object):
    '''
    User credentials for a specific wiki
    '''
    
    def __init__(self):
        '''
        construct me
        '''
        # set None values for all fields
        for field in WikiUser.getFields():
            self.__dict__[field]=None
    
    def getPassword(self):
        '''
        get my decrypted password
        
        Returns:
            str: the decrypted password for this user
        '''
        c=Crypt(self.cypher,20,self.salt)
        return c.decrypt(self.secret)
    
    def getWikiUrl(self):
        '''
        return the full url of this wiki
        
        Returns:
            str: the full url of this wiki
        '''
        url=f"{self.url}{self.scriptPath}"
        return url
    
    def interactiveSave(self,yes,filePath=None):
        '''
        save me
        '''
        fields=WikiUser.getFields(encrypted=False)
        for field in fields:
            if field not in self.__dict__ or self.__dict__[field] is None:
                inputMsg=f"{field}: "
                self.__dict__[field]=input(inputMsg)
        # encrypt
        self.encrypt()
        if not yes:
            answer=input("shall i store %s? yes/no y/n" % self)
            yes="y" in answer or "yes" in answer
        if yes:
            self.save(filePath)
    
    def encrypt(self):
        '''
        encrypt my clear text password
        '''
        crypt=Crypt.getRandomCrypt()
        self.secret=crypt.encrypt(self.password)
        self.cypher=crypt.cypher.decode()
        self.salt=crypt.salt.decode()
    
    def __str__(self):
        text=f"{self.user} {self.wikiId}"
        return text
    
    @staticmethod
    def getIniPath():
        home = str(Path.home())
        path=f"{home}/.mediawiki-japi"
        return path
    
    @staticmethod
    def iniFilePath(wikiId:str):
        user=getpass.getuser()
        iniPath=WikiUser.getIniPath()
        iniFilePath=f"{iniPath}/{user}_{wikiId}.ini"
        return iniFilePath
    
    @staticmethod
    def ofWikiId(wikiId:str,lenient=False):
        '''
        create a wikiUser for the given wikiId
        
        Args:
            wikiId(str): the wikiId of the user to be created
            lenient(bool): if True ignore parsing errors in the ini file
            
        Returns:
            WikiUser: the wikiUser for this wikiId
        '''
        path=WikiUser.iniFilePath(wikiId)
        try:
            config=WikiUser.readPropertyFile(path)
        except FileNotFoundError as _e:
            errMsg=f'the wiki with the wikiID "{wikiId}" does not have a corresponding configuration file ... you might want to create one with the wikiuser command'
            raise FileNotFoundError(errMsg)
        wikiUser=WikiUser.ofDict(config,lenient=lenient)
        return wikiUser
    
    def save(self,iniFilePath=None):
        '''
        save me to a propertyFile
        '''
        if iniFilePath is None:
            iniPath=WikiUser.getIniPath()
            if not isdir(iniPath):
                makedirs(iniPath)
            iniFilePath=WikiUser.iniFilePath(self.wikiId)
        
        iniFile=open(iniFilePath,"w")
        isodate=datetime.datetime.now().isoformat()
        template="""# Mediawiki JAPI credentials for %s
# created by py-3rdparty-mediawiki WikiUser at %s
"""
        content=template % (self.wikiId,isodate)
        for field in WikiUser.getFields():
            value=self.__dict__[field]
            if value is not None:
                content+="%s=%s\n" % (field,value) 
    
        iniFile.write(content)
        iniFile.close()
       
    @staticmethod
    def readPropertyFile(filepath, sep='=', comment_char='#'):
        """
        Read the file passed as parameter as a properties file.
        https://stackoverflow.com/a/31852401/1497139
        """
        props = {}
        with open(filepath, "rt") as f:
            for line in f:
                l = line.strip()
                if l and not l.startswith(comment_char):
                    key_value = l.split(sep)
                    key = key_value[0].strip()
                    value = sep.join(key_value[1:]).strip().strip('"') 
                    props[key] = value 
        return props
                    
    @staticmethod
    def getWikiUsers(lenient:bool=False):
        wikiUsers={}
        iniPath=WikiUser.getIniPath()
        if os.path.isdir(iniPath):
            with os.scandir(iniPath) as it:
                for entry in it:
                    if entry.name.endswith(".ini") and entry.is_file():
                        try:
                            config=WikiUser.readPropertyFile(entry.path)
                            wikiUser=WikiUser.ofDict(config,lenient=lenient)
                            wikiUsers[wikiUser.wikiId]=wikiUser
                        except Exception as ex:
                            print ("error in %s: %s" % (entry.path,str(ex)))
        return wikiUsers
    
    @staticmethod
    def getFields(encrypted=True):
        # copy fields
        fields=['email','scriptPath','user','url','version','wikiId']
        passwordFields=['cypher','secret','salt'] if encrypted else ['password']
        result=[]
        result.extend(fields)
        result.extend(passwordFields)
        return result
    
    @staticmethod
    def ofDict(userDict,encrypted=True,lenient=False):
        wikiUser=WikiUser()
        # fix http\: entries from Java created entries
        if 'url' in userDict and userDict['url'] is not None:
            userDict['url']=userDict['url'].replace("\:",":")
            
        for field in WikiUser.getFields(encrypted):
            if field in userDict:
                wikiUser.__dict__[field]=userDict[field]    
            else:
                if not lenient:
                    raise Exception(f"{field} missing")  
        if not encrypted:
            wikiUser.encrypt()   
        return wikiUser
    
__version__ = Version.version
__date__ = Version.date
__updated__ = Version.updated
DEBUG=False    

def main(argv=None): # IGNORE:C0111
    '''
    WikiUser credential handling
    '''

    if argv is None:
        argv = sys.argv[1:]
    
    program_name = os.path.basename(sys.argv[0]) 
    program_version = "v%s" % __version__
    program_build_date = str(__updated__)
    program_version_message = '%%(prog)s %s (%s)' % (program_version, program_build_date)
    program_shortdesc = "WikiUser credential handling"
    user_name="Wolfgang Fahl"
    
    program_license = '''%s

  Created by %s on %s.
  Copyright 2020-2022 Wolfgang Fahl. All rights reserved.

  Licensed under the Apache License 2.0
  http://www.apache.org/licenses/LICENSE-2.0

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
''' % (program_shortdesc,user_name, str(__date__))

    try:
        # Setup argument parser
        parser = ArgumentParser(description=program_license, formatter_class=RawDescriptionHelpFormatter)
        parser.add_argument("-d", "--debug", dest="debug",   action="count", help="set debug level [default: %(default)s]")
        parser.add_argument('-V', '--version', action='version', version=program_version_message)

        parser.add_argument("-e", "--email", dest="email", help="email of the user")
        parser.add_argument("-f", "--file", dest="filePath", help="ini-file path")
        parser.add_argument("-l", "--url", dest="url", help="url of the wiki")
        parser.add_argument("-s", "--scriptPath", dest="scriptPath", help="script path")
        parser.add_argument("-p", "--password", dest="password", help="password")
        parser.add_argument("-u", "--user", dest="user", help="os user id")
        parser.add_argument("-v", "--wikiVersion", dest="version", help="version of the wiki")
        parser.add_argument("-w", "--wikiId", dest="wikiId", help="wiki Id")
        parser.add_argument("-y", "--yes", dest="yes", action='store_true', help="immediately store without asking")
        # Process arguments
        args = parser.parse_args(argv)
        argsDict=vars(args)
        wikiuser=WikiUser.ofDict(argsDict,encrypted=False,lenient=True)
        wikiuser.interactiveSave(args.yes,args.filePath)
        
    except KeyboardInterrupt:
        ### handle keyboard interrupt ###
        return 1
    except Exception as e:
        if DEBUG:
            raise(e)
        indent = len(program_name) * " "
        sys.stderr.write(program_name + ": " + repr(e) + "\n")
        sys.stderr.write(indent + "  for help use --help")
        return 2

if __name__ == '__main__':
    if DEBUG:
        sys.argv.append("-d")
    sys.exit(main())


'''
Created on 31.01.2021

@author: wf
'''
from lodstorage.query import Query
from jinja2 import Template,Environment
import html
import re

class Code(object):
    '''
    a piece of code
    '''
    
    def __init__(self, name:str,text:str,lang:str='python'):
        '''
        construct me from the given text and language
        '''
        self.name=name
        self.text=text
        self.lang=lang
        
    def execute(self,context):
        '''
        https://stackoverflow.com/questions/701802/how-do-i-execute-a-string-containing-python-code-in-python
        https://stackoverflow.com/questions/436198/what-is-an-alternative-to-execfile-in-python-3
        https://stackoverflow.com/questions/2220699/whats-the-difference-between-eval-exec-and-compile
        '''
        if self.lang == "jinja":
            self.executeTemplate(context)
        else:
            exec(self.text)
        pass

    def executeTemplate(self, context):
        """
        Renders the jinja-template with the query results placed in the given context and stores the result as wiki page.
        The name of the wiki page is either given through the template with set parameters.
        E.g.:
            {% set pagetitle = ""%}
            {% set pagetitle_prefix = "List of "%}
        If the pagetitle is empty, like in the example, the name variable of the query results is used as pagetitle.
        The pagetitle_prefix is added in all cases, if not defined a empty string is added.
        Assumption: self.text is a jinja template
        """
        getAttribute = lambda template, name : re.search("\{% *set +" + name + " += *['\"](?P<name>.*)['\"] *%\}",template).group("name")
        raw_template = LambdaAction.unescapeHTML(self.text)
        template = Template(raw_template)
        title = getAttribute(raw_template, "pagetitle")
        pagetitle_prefix = getAttribute(raw_template, "pagetitle_prefix")
        if not pagetitle_prefix:
            pagetitle_prefix=""
        user = context["wikiclient"].wikiUser.user
        for row in context["rows"]:
            page_content = template.render({"row": row, "smw": context["smw"]})
            if not title:
                if not row['name']:
                    raise ValueError(f"Can't save wikipage without a title. Either provide a page title in the "
                                     f"template or provide the name variable for each query result")
                pagetitle = pagetitle_prefix + row['name']
            else:
                pagetitle = pagetitle_prefix + title
            page = context['wikiclient'].getPage(pagetitle)
            # ToDo
            page.edit(page_content, f"Created with the template [[{self.name}]] by {user} through LambdaActions")


class LambdaAction(object):
    '''
    a lambda action
    '''

    def __init__(self, name:str,query:Query,code:Code):
        '''
        Constructor
        '''
        self.name=name
        self.query=query
        self.code=code
        
    def executeQuery(self,context):
        rows=None
        if self.query.lang == "sql":
            if "sqlDB" in context:
                db=context["sqlDB"]
                rows=db.query(self.query.query)
        elif self.query.lang == "sparql":
            # ToDo
            pass
        elif self.query.lang == "smw":
            if "smw" in context:
                smw = context["smw"]
                query = LambdaAction.unescapeHTML(self.query.query)
                query_results = smw.query(query)
                rows = list(query_results.values())
        else:
            print(f"Queries of type {self.query.lang} are currently not supported by LambdaActions.")
        context["rows"] = rows
        return rows
            
    def getMessage(self,context):
        message=None
        if 'result' in context:
            result=context['result']
            if 'message' in result:
                message=result["message"]
        return message
        
    def execute(self,context):
        '''
        run my query and feed the result into the given code
        
        Args:
            context(dict): a dictionary for the exchange of parameters
        '''
        self.executeQuery(context)
        self.code.execute(context)

    @staticmethod
    def unescapeHTML(value:str):
        """
        Unescapes received html value and removes html tags.
        Replaces:
            <br /> -> "\n"
            <pre> -> ""
        Args:
            value(str): html encoded string
        Returns:
            Returns the received value but without the html tags and unescaped.
        """
        if value.startswith("<pre>"):
            return html.unescape(value).replace("<br />", "\n")[5:][:-6]
        return value




'''
Created on 2020-11-05

@author: wf
'''

class Wiki(object):
    '''
    common interface for WikiBot and WikiClient
    '''

    def __init__(self,wikiUser,debug=False):
        '''
        Constructor
        
        Args:
            wikiUser(WikiUser): the wiki user to initialize me for
        '''
        self.wikiUser=wikiUser
        self.debug=debug
        
    def __str__(self):
        '''
        return a string representation of myself
        '''
        wu=self.wikiUser
        botType=type(self).__name__
        text="%20s(%10s): %15s %s" % (wu.wikiId,botType,wu.user,wu.url)    
        return text
        


'''
Created on 02.02.2021

@author: wf
'''
from wikibot.lambda_action import Code,LambdaAction
from lodstorage.query import Query

class WikiAction(object):
    '''
    perform an action on the given semantic media wiki
    '''

    def __init__(self, smw, debug=False):
        '''
        Constructor
        '''
        self.smw=smw
        self.debug=debug
        self.sourceCodeResult=self.getSourceCodes()
        
    def getSourceCodes(self):
        ask="""{{#ask: [[Concept:Sourcecode]]
|mainlabel=Sourcecode
| ?Sourcecode id = id
| ?Sourcecode lang = lang
| ?Sourcecode text = text
| ?Sourcecode author = author
| ?Sourcecode since = since
| ?Sourcecode url = url
}}"""   
        result=self.smw.query(ask)
        return result
    
    def getLambdaAction(self,name,queryName,actionName):
        '''
        get an action from the result with the given query and actionName
        '''
        if self.debug:
            print("lambdaAction with query: %s and action: %s" % (queryName,actionName))
        qCode=self.sourceCodeResult[queryName]
        query=Query(name=qCode['id'],query=qCode['text'],lang=qCode['lang'])
        sCode=self.sourceCodeResult[actionName]
        code=Code(name=sCode['id'],text=sCode['text'],lang=sCode['lang'])
        action=LambdaAction(name,query=query,code=code)
        return action
          
        
    
        
        
        


'''
Created on 2020-12-20

'''


class Selector:
    '''
    selector class
    '''

    def __init__(self, items):
        """
            Constructor
        """
        import tkinter as tk 
        self.items = items
        self.items = list(map(str, self.items))
        self.var = dict()
        self.count = 1
        self.checkvar = tk.IntVar(value=1)
        self.label_test = tk.StringVar()
        self.label_test.set("Select None")
        self.quitProgram = False

    def createWindow(self, root, action, title, description):
        """
            create the Window for the selection list
            Args:
               root(tk.App Object): tk.App opened object
               action(str): Type of Action
               title(str): Title of Window
               description(str): Description of Task to do
           Returns:
               None
        """
        import tkinter as tk 
        # Title
        rowCounter = 0
        root.title(title)    #set window title to given title string
        root.resizable(1, 1) #set so the window can be resized by user

        # finding optimal window size
        items_by_len = sorted(self.items, key=len)
        longest_string = items_by_len[-1]

        # Setting Message to description string
        desc = tk.Message(root, text=description)
        desc.bind("<Configure>", lambda e: desc.configure(width=e.width - 10))
        desc.pack(side="top", fill=tk.X)

        # Frame creation for Listbox
        frameList = tk.Frame(root)

        items = tk.StringVar(root)
        items.set(self.items)
        listbox = tk.Listbox(frameList, listvariable=items, selectmode="multiple", width=len(longest_string)+5)
        listbox.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        listbox.select_set(0, tk.END)

        # Frame creation for select all/none checkbox
        frameSelect = tk.Frame(root)
        frameSelect.pack(side=tk.TOP, fill=tk.BOTH)
        frameList.pack(fill=tk.BOTH, expand=True)

        # label and checkbutton creation
        label = tk.Label(frameSelect, textvariable=self.label_test)
        check = tk.Checkbutton(frameSelect, text="", variable=self.checkvar,
                            command=lambda: self.select_all(listbox))
        check.pack(side=tk.LEFT, anchor="sw")
        label.pack(side=tk.LEFT, anchor="sw")

        # Scrollbar binding and creation
        scrollbar = tk.Scrollbar(frameList)
        scrollbar.pack(side=tk.RIGHT, fill=tk.BOTH)
        listbox.config(yscrollcommand=scrollbar.set)
        scrollbar.config(command=listbox.yview)
        listbox.bind("<<ListboxSelect>>", self.updateCheck)

        #Frame for Action and Quit Buttons
        frameControl = tk.Frame(root)
        frameControl.pack(side=tk.BOTTOM, fill=tk.X)

        # Action and Quit button creation
        actionBtn = tk.Button(frameControl, text=action, bg='green', fg='white', command=lambda: self.updatePages(root, listbox))
        actionBtn.pack(side=tk.LEFT, anchor="sw")
        quitBtn = tk.Button(frameControl, text="Quit", bg='red', fg='white', command=lambda: self.quitSelector(root))
        quitBtn.pack(side=tk.RIGHT, anchor="se")

        #To destroy window and exit program if cross button is pressed
        root.protocol("WM_DELETE_WINDOW", lambda: self.quitSelector(root))

    def updateCheck(self, event):
        """
            Helper function to change state of select all checkbox
            Returns:
                event(Tk event Object): All registered events on Tkinter
        """
        if len(event.widget.curselection()) < len(self.items): #check if any item is deselected
            self.label_test.set("Select All")
            self.checkvar.set(0)
        elif len(event.widget.curselection()) == len(self.items):
            self.label_test.set("Select None")
            self.checkvar.set(1)

    def select_all(self, listbox):
        """
            Button helper function to select all list items
            Returns:
                listbox(Tk Listbox Object): listbox to update to all items
        """
        import tkinter as tk 
        if self.checkvar.get():
            listbox.select_set(0, tk.END)
            self.label_test.set("Select None")
        else:
            listbox.select_clear(0, tk.END)
            self.label_test.set("Select All")

    def updatePages(self, root, listbox):
        """
            Update function to remove unselected items from list
            Args:
                root(tk.App Object): tk.App opened object
                listbox(TK listbox Object): listbox to update
            Returns:
                None
        """
        self.items = [listbox.get(idx) for idx in listbox.curselection()] #Remove unselected items from list
        root.destroy()

    def quitSelector(self, root):
        """
            Quit the python program when Quit Button is pressed.
            Args:
                root(tk.App Object): tk.App opened object
            Returns:
                None
        """
        root.destroy()
        self.items = 'Q'


    def getUpdatedPages(self):
        """
           Getter function for the class variable items
           Returns:
               items(list): List of pages selected in GUI by user
       """
        return self.items


    @staticmethod
    def select(selectionList, action="Select", title="Selection", description=""):
        """
        Creates a GUI in which the user can select a subset of the provided selectionList.
        The user can quit the selection resulting in the return of an empty list.
        :param selectionList:
        :param action: name of the action the selection performs. Default is select
        :param title: title of the created window
        :param description: Instructions for the user (consequence of the selection)
        :return: user selected subset of the given selectionList
        """
        import tkinter as tk 
        root = tk.Tk()
        GUI = Selector(selectionList)   # Tkinter Object creation
        GUI.createWindow(root, action, title, description)  # create Window with given parameters
        root.mainloop()     # Run GUI loop
        selectionList = GUI.getUpdatedPages()   # Get selected items
        return selectionList





'''
Created on 2020-03-24

@author: wf
'''
import sys
from urllib.parse import urlparse
try:
    import pywikibot
    from pywikibot.data.api import LoginManager
except Exception as ex:
    # ignore pywikibot setup problems?
    print("pywikibot config issue see https://phabricator.wikimedia.org/T278076")
    print("%s" % str(ex), file=sys.stderr) 
    pass
import re
from os.path import isfile 
from wikibot.wikiuser import WikiUser
from wikibot.wiki import Wiki

class WikiBot(Wiki):
    '''
    WikiBot
    '''      
    @staticmethod
    def getBots(limit=None,name=None,valueExpr=None):
        bots={}
        wikiUsers=WikiUser.getWikiUsers().values()
        for wikiUser in wikiUsers:
            selected=True
            if name is not None:
                value=wikiUser.__dict__[name]
                found=re.search(valueExpr,value)
                selected=found is not None
            if selected:
                wikibot=WikiBot(wikiUser)
                bots[wikiUser.wikiId]=wikibot
                if limit is not None and len(bots)>=limit:
                    break
        return bots
    
    @staticmethod
    def ofWikiId(wikiId,lenient=True,debug=False):
        wikiUser=WikiUser.ofWikiId(wikiId,lenient=lenient)
        wikibot=WikiBot(wikiUser,debug=debug)
        return wikibot
    
    @staticmethod
    def ofWikiUser(wikiUser):
        wikibot=WikiBot(wikiUser)
        return wikibot
        
    def __init__(self,wikiUser,debug:bool=False,withLogin:bool=False,maxRetries=2):
        '''
        Constructor
        
        Args:
            wikiUser(WikiUser): the wiki user to initialize me for
            debug(bool): True if debug mode should be on
            withLogin(bool): True if init should automatically login
        '''
        pywikibot.config.max_retries=maxRetries
        super(WikiBot,self).__init__(wikiUser,debug) 
        self.family=wikiUser.wikiId.replace("-","").replace("_","")
        self.url=wikiUser.url.replace("\\:",":")
        if not self.url:
            raise Exception("url is missing for %s" % wikiUser.wikiId)
        
        self.scriptPath=wikiUser.scriptPath
        self.version=wikiUser.version
        o=urlparse(self.url)
        self.scheme=o.scheme
        self.netloc=o.netloc
        self.scriptPath=o.path+self.scriptPath      
        self.checkFamily()
        if withLogin:
            self.login()
        
    def register_family_file(self,familyName:str,famfile:str):
        '''
        register the family file
        
        Args:
            family(str): the familyName to register
            famfile(str): the path to the family file
        '''
        # deprecated code
        #config2.register_family_file(familyName, famfile)
        pywikibot.config.family_files[familyName]=famfile  
  
        
    def checkFamily(self):
        '''
        check if a family file exists and if not create it
        
        '''
        iniFile=WikiUser.iniFilePath(self.wikiUser.wikiId)
        famfile=iniFile.replace(".ini",".py")
        if not isfile(famfile):
            print("creating family file %s" % famfile)
            template='''# -*- coding: utf-8  -*-
from pywikibot import family

class Family(family.Family):
    name = '%s'
    langs = {
        'en': '%s',
    }
    def scriptpath(self, code):
       return '%s'
       
    def isPublic(self):
        return %s   
        
    def version(self, code):
        return "%s"  # The MediaWiki version used. Very important in most cases. (contrary to documentation)   

    def protocol(self, code):
       return '%s'
'''         
            mw_version=self.wikiUser.version.lower().replace("mediawiki ","")
            ispublic='False' if self.wikiUser.user is not None else 'True'
            code=template % (self.family,self.netloc,self.scriptPath,ispublic,mw_version,self.scheme)
            with open(famfile,"w") as py_file:
                py_file.write(code)
        self.register_family_file(self.family,famfile)
        if self.wikiUser.user:
            pywikibot.config.usernames[self.family]['en'] = self.wikiUser.user
        #config2.authenticate[self.netloc] = (self.user,self.getPassword())
        self.site=pywikibot.Site('en',self.family)  
        
    def login(self):
        if self.wikiUser.user:
            # needs patch as outlined in https://phabricator.wikimedia.org/T248471
            #self.site.login(password=self.wikiUser.getPassword())
            lm = LoginManager(password=self.wikiUser.getPassword(), site=self.site, user=self.wikiUser.user)
            lm.login()
        else:
            raise Exception("wikiUser is not set")
            
    def getWikiMarkup(self,pageTitle):
        '''
        get the wiki markup code (text) for the given page Title
        
        Args:
            pageTitle(str): the title of the page to retrieve
            
        Returns:
            str: the wiki markup code for the page
        '''
        page=self.getPage(pageTitle)
        markup=page.text
        return markup
            
    def getHtml(self,pageTitle):
        '''
        get the HTML code for the given page Title
        
        Args:
            pageTitle(str): the title of the page to retrieve
            
        Returns:
            str: the rendered HTML code for the page
        '''
        page=self.getPage(pageTitle)
        html=page.get_parsed_page()
        return html
        
    def getPage(self,pageTitle):
        ''' get the page with the given title
        Args:
            pageTitle(str): the title of the page to retrieve
        Returns:
            Page: the wikibot page for the given pageTitle
        '''
        page = pywikibot.Page(self.site, pageTitle)  
        return page             
    
    def savePage(self,pageTitle,pageContent,pageSummary):
        ''' save a page with the given pageTitle, pageContent and pageSummary'''
        newPage=self.getPage(pageTitle)
        newPage.text=pageContent
        newPage.save(pageSummary)



'''
Created on 25.03.2020

@author: wf
'''
from Crypto.Hash import MD5
from Crypto.Cipher import DES
import base64
import secrets
import string
import os
import warnings

class Crypt(object):
    '''
    Python implementation of PBEWithMD5AndDES
    see
    https://github.com/binsgit/PBEWithMD5AndDES
    and
    https://gist.github.com/rohitshampur/da5f79c34260150aafc1
    
    converted to class
'''
    def __init__(self,cypher,iterations=20,salt=None):
        ''' construct me with the given cypher iterations and salt 
        '''
        # avoid annoying
        #  /opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/Crypto/Cipher/blockalgo.py:141: DeprecationWarning: PY_SSIZE_T_CLEAN will be required for '#' formats
        warnings.filterwarnings("ignore", category=DeprecationWarning) 
 
        self.cypher=cypher.encode('utf-8')
        self.iterations=iterations
        if salt is None:
            self.salt=os.urandom(8)
        else:
            self.salt=salt.encode('utf-8')      
        pass
    
    @staticmethod
    def getRandomString(rlen=32):
        #https://docs.python.org/3/library/secrets.html
        alphabet = string.ascii_letters + string.digits
        rstring = ''.join(secrets.choice(alphabet) for i in range(rlen))
        return rstring
    
    @staticmethod
    def getRandomCrypt(cypherLen=32):
        cypher=Crypt.getRandomString(cypherLen)
        salt=Crypt.getRandomString(8)
        crypt=Crypt(cypher,salt=salt)
        return crypt;
    
    def getCrypt(self):
        '''
        get my DES crypt
        '''
        hasher = MD5.new()
        hasher.update(self.cypher)
        hasher.update(self.salt)
        result = hasher.digest()
    
        # iterate over hashes
        for _i in range(1, self.iterations):
            hasher = MD5.new()
            hasher.update(result)
            result = hasher.digest()
        key=result[:8]
        # initialization vector
        iv=result[8:16]
        return DES.new(key, DES.MODE_CBC, iv)
    
    def encrypt(self,msg):
        '''
        encrypt the given message
        '''
        plaintext_to_encrypt = msg
        # Pad plaintext per RFC 2898 Section 6.1
        padding = 8 - len(plaintext_to_encrypt) % 8
        plaintext_to_encrypt += chr(padding) * padding
        encoder = self.getCrypt()
        encrypted = encoder.encrypt(plaintext_to_encrypt.encode('utf-8'))
        b64enc=base64.b64encode(encrypted).decode('utf-8')
        return b64enc

    def decrypt(self,encoded):
        '''
        decrypt the given message 
        '''
        enc = base64.b64decode(encoded)
        decoder = self.getCrypt()
        decryptedb = decoder.decrypt(enc)
        decrypted=decryptedb.decode('utf-8')
        return decrypted.rstrip('\2,\1,\3,\4,\5,\6,\7,\0,\b')
    

if __name__ == "__main__":
    msg = "hello, world"
    cypher = "mycypher"
    c=Crypt(cypher)
    s = c.encrypt(msg)
    print (s)
    d=c.decrypt(s)
    print (d)


'''
Created on 2020-05-29

@author: wf
'''
import re
import sys
from datetime import datetime
from urllib.parse import unquote
import pywikibot

class PrintRequest(object):
    debug=False
    """
    construct the given print request
    see https://www.semantic-mediawiki.org/wiki/Serialization_(JSON)
    :ivar smw: SMW context for this printrequest
    :ivar label: the label of the printrequest
    :ivar key: 
    :ivar redi: 
    :ivar typeid: 
    :ivar mode:  
    :ivar format: 
    """
    def __init__(self,smw,record):
        '''
        construct me from the given record
        Args:
            smw(SMW): the SemanticMediaWiki context of this PrintRequest
            record(dict): the dict derived from the printrequest json serialization  
        '''
        self.smw=smw
        self.debug=PrintRequest.debug
        if self.debug:
            print(record)
        self.label=record['label']
        self.key=record['key']
        self.redi=record['redi']
        self.typeid=record['typeid']
        self.mode=int(record['mode'])
        if 'format' in record:
            self.format=record['format']
        else:
            self.format=None   
            
    def deserializeSingle(self,value):
        """ deserialize a single value 
        Args:
            value(object): the value to be deserialized according to the typeid
            
        Returns:
            the deserialized value
        """
        # FIXME complete list of types according to
        # https://www.semantic-mediawiki.org/wiki/Help:API:ask  
        # Page https://www.semantic-mediawiki.org/wiki/Help:API:ask/Page    
        if self.typeid=="_wpg":  
            value=value["fulltext"]
            if value:
                value=unquote(value)
            pass  
        # Text https://www.semantic-mediawiki.org/wiki/Help:API:ask/Text
        elif self.typeid=="_txt":
            pass
        elif self.typeid=="_qty":    
            pass
        elif self.typeid=="_num":
            value=int(value)            
        elif self.typeid=="_dat":
            if 'timestamp' in value:
                ts=int(value['timestamp'])
                try:
                    value=datetime.utcfromtimestamp(ts)
                    #  print (date.strftime('%Y-%m-%d %H:%M:%S'))
                except ValueError as ve:
                    if self.debug:
                        print("Warning timestamp %d is invalid: %s" % (ts,str(ve)))
                    pass
            else:
                # ignore faulty values
                if self.debug:
                    print("Warning: timestamp missing for value")
                pass
        elif self.typeid=="_eid":
            pass
        else:   
            pass
        return value
            
    def deserialize(self,result):
        """ deserialize the given result record
        Args:
            result(dict): a single result record dict from the deserialiation of the ask query
        Returns:    
            object: a single deserialized value according to my typeid   
        """
        po=result['printouts']
        if self.label in po:
            value=po[self.label]
        else:
            value=result
        if isinstance(value,list):
            valueList=[]
            for valueItem in value:
                valueList.append(self.deserializeSingle(valueItem))
            # handle lists
            # empty lists => None
            if len(valueList)==0:
                value=None    
            # lists with one value -> return the item (this unfortunately removes the list property of the value)   
            elif len(valueList)==1:
                value=valueList[0]
            # only if there is a "real" list return it    
            else:        
                value=valueList  
        else:
            value=self.deserializeSingle(value)                           
        
        if PrintRequest.debug:
            print ("%s(%s)='%s'" % (self.label,self.typeid,value))  
        return value    
            
    def __repr__(self):
        text="PrintRequest(label='%s' key='%s' redi='%s' typeid='%s' mode=%d format='%s')" % (self.label,self.key,self.redi,self.typeid,self.mode,self.format)
        return text
    
class SplitClause():
    '''
    Query parameter to be used for splitting e.g. Modification date, Creation Date, could be potentially
    any parameter that is ordered and countable
    Currently we assume a parameter of type Date and use Modification date by default
    '''
    
    def __init__(self,name="Modification date",label="mdate"):
        '''
        construct me
        '''
        self.name=name
        self.label=label
        
    def queryBounds(self,lowerBound,upperBound)->str:
        '''
        get the query bounds
        e.g. [[Modification date::>=2021-01-01T12:00

        Args:
            lowerBound(datetime): start of the time interval
            upperBound(datetime): end of the time interval

        Returns:
            Returns the SMW ask part for the boundary
        '''
        result=f"[[{self.name}:: >={lowerBound.isoformat()}]]|[[{self.name}:: <={upperBound.isoformat()}]]"
        return result

    def getFirst(self)->str:
        '''
        get the first element
        '''
        result=f"?{self.name}={self.label}|sort={self.name}|limit=1"
        return result
    
    def deserialize(self,values):
        '''
        deserialize my query result
        '''
        # dict: {
        #    'Property:Foaf:knows': {
        #       '': 'Property:Foaf:knows', 
        #       'Modification date': datetime.datetime(2020, 11, 28, 17, 40, 36)
        #    }
        # }
        date=None
        vlist=list(values)
        if len(vlist)>0:
            innerValue = vlist[0]
            if self.label in innerValue:
                date=innerValue[self.label]
        return date
        

class SMW(object):
    '''
    Semantic MediaWiki Access e.g. for ask API
    see
    * https://www.semantic-mediawiki.org/wiki/Help:API
    * https://www.semantic-mediawiki.org/wiki/Serialization_(JSON)
    * https://www.semantic-mediawiki.org/wiki/Help:API:askargs
    
    adapted from Java SimpleGraph Module 
    https://github.com/BITPlan/com.bitplan.simplegraph/blob/master/simplegraph-smw/src/main/java/com/bitplan/simplegraph/smw/SmwSystem.java
    :ivar site: the pywikibot site to use for requests
    :ivar prefix: the path prefix for this site e.g. /wiki/
    '''
    def __init__(self, site=None,prefix="/",showProgress=False, queryDivision=1,debug=False):
        '''
        Constructor
        Args:
            site: the site to use (optional)
            showProgess(bool): if progress should be shown
            queryDivision(int): Defines the number of subintervals the query is divided into (must be greater equal 1)
            debug(bool): if debugging should be activated - default: False
        '''
        self.site=site
        self.prefix=prefix
        self.showProgress=showProgress
        self.queryDivision=queryDivision
        self.splitClause=SplitClause()
        self.debug=debug
    
    def deserialize(self,rawresult) -> dict:
        """ deserialize the given rawresult according to 
        https://www.semantic-mediawiki.org/wiki/Serialization_(JSON)
        
        Args:
            rawresult(dict): contains printrequests and results which need to be merged
            
        Returns:
            dict: query mainlabel (usually pageTitle) mapped to the corresponding dict of printrequests with label
        """
        resultDict={}
        if rawresult is None:
            return resultDict
        if not 'query' in rawresult:
            raise Exception("invalid query result - 'query' missing")
        query=rawresult['query']
        if not 'printrequests' in query:
            raise Exception("invalid query result - 'printrequests' missing")
        printrequests=query['printrequests']
        if not 'results' in query:
            raise Exception("invalid query result - 'results' missing")
        results=query['results']
        prdict={}
        for record in printrequests:
            pr=PrintRequest(self,record)
            prdict[pr.label]=pr
        
        if results:
            for key in results.keys():
                record=results[key]
                recordDict={}
                for label in prdict.keys():
                    pr=prdict[label]
                    recordDict[label]=pr.deserialize(record)
                resultDict[key]=recordDict
        return resultDict
    
    def fixAsk(self, ask:str):
        """
        fix an ask String to be usable for the API
        Args:
            ask: - a "normal" ask query

        Returns:
             the fixed asked query
        """
        # ^\\s*\\{\\{
        # remove {{ with surrounding white space at beginning
        fixedAsk = re.sub(r"^\s*\{\{", "",ask)
        # remove #ask:
        fixedAsk = re.sub(r"#ask:", "",fixedAsk)
        # remove }} with surrounding white space at end
        fixedAsk = re.sub(r"\}\}\s*$", "",fixedAsk)
        # split by lines (with side effect to remove newlines)
        parts = fixedAsk.split(r"\n")
        fixedAsk = ""
        for part in parts:
            #  remove whitespace around part
            part = part.strip()
            # remove whitespace around pipe sign
            part = re.sub(r"\s*\|\s*", "|",part)
            # remove whitespace around assignment =
            part = re.sub(r"\s*=\s*", "=",part)
            # remove whitespace in query parts
            part = re.sub(r"\]\s*\[", "][",part)
            fixedAsk = fixedAsk + part
        return fixedAsk
    
    def getConcept(self,ask):
        """ get the concept from the given ask query"""
        m=re.search(r"\[\[Concept:(.+?)\]\]",ask)
        if m:
            return m.groups()[0]
        else:
            return None

    argumentRegex = staticmethod(lambda arg: r"\| *" + arg + r" *= *\d+(?=\s|\||$)")

    @staticmethod
    def getOuterMostArgumentValueOfQuery(argument, query):
        """
        Extracts the integer value of the given argument from the given query
        Args:
            argument(string): Argument that should be extracted
            query(string): smw query where the given argument is assumed

        Returns:
            Returns integer value of the given argument in the given query.
            If the argument occurs multiple times the last one is returned.
            If it does not occur return None.
        """
        if not argument or not query:
            return None
        args = re.compile(SMW.argumentRegex(argument), re.IGNORECASE).findall(query)
        if not args:
            return None
        return int(re.compile("[0-9]+").search(args[-1]).group())


class SMWClient(SMW):
    '''
    Semantic MediaWiki access using mw client library
    '''
    
    def __init__(self, site=None,prefix="/",showProgress=False, queryDivision=1,debug=False):
        super(SMWClient,self).__init__(site,prefix,showProgress=showProgress, queryDivision=queryDivision,debug=debug)
        pass
    
    def info(self):
        """ see https://www.semantic-mediawiki.org/wiki/Help:API:smwinfo"""
        results = self.site.raw_api('smwinfo', http_method='GET')
        self.site.handle_api_result(results)  # raises APIError on error
        
        return results
    
    def ask(self, query:str, title:str=None, limit:int=None):
        """
        Ask a query against Semantic MediaWiki.

        API doc: https://semantic-mediawiki.org/wiki/Ask_API

        The query is devided into multiple subqueries if the results of the query exeed the defined threshold.
        If this happens the query is executed multiple times to retrieve all results without passing the threshold.


        Args:
            query(str): SMW ask query to be executed
            title(str): title of query (optional)
            limit(int): the maximum number of results to be returned - 
                please note that SMW configuration will also limit results on the server side
        
        Returns:
            Generator for retrieving all search results, with each answer as a dictionary.
            If the query is invalid, an APIError is raised. A valid query with zero
            results will not raise any error.

        Examples:

            >>> query = "[[Category:my cat]]|[[Has name::a name]]|?Has property"
            >>> for answer in site.ask(query):
            >>>     for title, data in answer.items()
            >>>         print(title)
            >>>         print(data)
        """
        #kwargs = {}
        #if title is None:
        #    kwargs['title'] = title

        if limit is None:
            # if limit is not defined (via cmd-line), check if defined in query
            limit = SMW.getOuterMostArgumentValueOfQuery("limit", query)
        results = None
        if self.queryDivision == 1:
            try:
                results = self.askForAllResults(query, limit)
            except QueryResultSizeExceedException as e:
                print(e)
                results = e.getResults()
        else:
            results = self.askPartitionQuery(query, limit)
        for res in results:
            yield res

    def askPartitionQuery(self, query, limit=None):
        """
        Splits the query into multiple subqueries by determining the 'modification date' interval in which all results
        lie. This interval is then divided into subintervals. The number of subintervals is defined by the user via
        commandline. The results of all subqueries are then returned.
        Args:
            query(string): the SMW inline query to be send via api
            limit(string): limit of the query
        Returns:
            All results of the given query.
        """
        (start, end) = self.getBoundariesOfQuery(query)
        results = []
        if start is None or end is None:
            return results
        if self.showProgress:
            print(f"Start: {start}, End: {end}", file=sys.stderr, flush=True)
        numIntervals = self.queryDivision
        calcIntervalBound = lambda start, n: (start + n * lenSubinterval).replace(microsecond=0)
        calcLimit = lambda limit, numRes: None if limit is None else limit - numResults
        done = False
        numResults = 0
        while not done:
            lenSubinterval = (end - start) / numIntervals
            for n in range(numIntervals):
                if self.showProgress:
                    print(f"Query {n+1}/{numIntervals}:")
                tempLowerBound = calcIntervalBound(start,n)
                tempUpperBound = calcIntervalBound(start,n+1) if (n+1 < numIntervals) else end
                queryBounds = self.splitClause.queryBounds(tempLowerBound, tempUpperBound)
                queryParam=f"{query}|{queryBounds}"
                try:
                    tempRes = self.askForAllResults(queryParam, calcLimit(limit, numResults))
                    if tempRes is not None:
                        for res in tempRes:
                            results.append(res)
                            query_field=res.get("query")
                            if query_field is not None:
                                meta_field=query_field.get("meta")
                                if meta_field is not None:
                                    numResults += int(meta_field.get("count"))
                except QueryResultSizeExceedException as e:
                    # Too many results for current subinterval n -> print error and return the results upto this point
                    print(e)
                    if e.getResults():
                        for res in e.getResults():
                            results.append(res)
                    numResults = 0
                    break
                if limit is not None and limit <= numResults:
                    if self.showProgress:
                        print(f"Defined limit of {limit} reached - ending querying")
                    done = True
                    break
            done=True
        return results
    
    def getTimeStampBoundary(self,queryparam,order):
        '''
        query according to a DATE e.g. MODIFICATION_DATE in the given order
        
        Args:
           
        '''
        queryparamBoundary = f"{queryparam}|order={order}"
        resultsBoundary = self.site.raw_api('ask', query=queryparamBoundary, http_method='GET')
        self.site.handle_api_result(resultsBoundary)
        deserializedResult=self.deserialize(resultsBoundary)
       
        deserializedValues=deserializedResult.values()
        date=self.splitClause.deserialize(deserializedValues)
        return date

    def getBoundariesOfQuery(self, query):
        """
        Retrieves the time interval, lower and upper bound, based on the modification date in which the results of the
        given query lie.
        Args:
            query(string): the SMW inline query to be send via api
        Returns:
            (Datetime, Datetime): Returns the time interval (based on modification date) in which all results of the
            query lie potentially the start end end might be None if an error occured or the input is invalid
        """
        first=self.splitClause.getFirst()
        queryparam = f"{query}|{first}"
        start=self.getTimeStampBoundary(queryparam,'asc')
        end=self.getTimeStampBoundary(queryparam,'desc')
        return (start,end)
    
    def askForAllResults(self, query, limit=None, kwargs={}):
        """
        Executes the query until all results are received of the given limit is reached.
        If the SMW results limit is reached before all results are retrieved the QueryResultSizeExceedException is raised.
        Args:
            query(string): the SMW inline query to be send via api
            limit(int): limit for the query results, None (default) for all results
            kwargs:
        Returns:
            query results
        Raises:
            QueryResultSizeExceedException: Raised if not all results can be retrieved
        """
        endShowProgress = lambda showProgress, c: print("\n" if not c % 80 == 0 else "") if showProgress else None
        offset = 0
        done = False
        count = 0
        res = []
        while not done:
            count += 1
            if self.showProgress:
                sep = "\n" if count % 80 == 0 else ""
                print(".", end=sep, flush=True)

            queryParam = u'{query}|offset={offset}'.format(query=query, offset=offset)
            if limit is not None:
                queryParam += "|limit={limit}".format(limit=limit)
            # print(f"QueryPram: {queryParam}")   #debug purposes
            results = self.site.raw_api('ask', query=queryParam, http_method='GET', **kwargs)
            self.site.handle_api_result(results)  # raises APIError on error
            continueOffset = results.get('query-continue-offset')
            if continueOffset is None:
                done = True
            else:
                if limit is not None and continueOffset >= limit:
                    done = True
                elif (results.get('query') is not None and not results.get('query').get('results')) or continueOffset < offset:
                    # contine-offset is set but result is empty
                    endShowProgress(self.showProgress, count)
                    res.append(results)
                    raise QueryResultSizeExceedException(result=res)
                if continueOffset < offset:
                    done = True
            offset = continueOffset
            res.append(results)
        endShowProgress(self.showProgress, count)
        return res
    
    def rawquery(self,askQuery,title=None,limit=None):
        '''
        run the given askQuery and return the raw result
        
        Args:
            askQuery(string): the SMW inline query to be send via api
            title(string): the title (if any)
            limit(int): the maximum number of records to be retrieved (if any)
            
        Returns:
            dict: the raw query result as returned by the ask API
        '''
        fixedAsk=self.fixAsk(askQuery)
        result=None
        for singleResult in self.ask(fixedAsk, title, limit):
            if result is None:
                result=singleResult
            else:
                singleResults=None
                if 'query' in singleResult:
                    if 'results' in singleResult['query']:
                        singleResults=singleResult['query']['results']
                if singleResults is not None:
                    if 'query' in result:
                        if 'results' in result['query']:
                            results=result['query']['results']
                            results.update(singleResults)
                        else:
                            result['query']['results']=singleResults
                    else:
                        result['query']={}
                        result['query']['results'] = singleResults
        return result
        
    def query(self,askQuery:str,title:str=None,limit:int=None) -> dict:
        '''
        run query and return list of Dicts
        
        Args:
            askQuery(string): the SMW inline query to be send via api
            title(string): the title (if any)
            limit(int): the maximum number of records to be retrieved (if any)
            
        Return:
            dict: mainlabel as key and value is a dict of the associated property values
        '''
        rawresult=self.rawquery(askQuery, title, limit)
        resultDict=self.deserialize(rawresult)
        return resultDict

    def updateProgress(self, count):
        if self.showProgress:
            sep = "\n" if count % 80 == 0 else ""
            print(".", end=sep, flush=True)
    
class SMWBot(SMW):
    '''
    Semantic MediaWiki access using pywikibot library
    '''
    def __init__(self, site=None,prefix="/",showProgress=False,debug=False):
        '''
        constructor
        
        Args:
            site:
            prefix(str): the prefix to use
            showProgress(bool): show progress if true
            debug(bool): set debugging mode if true
        '''
        super(SMWBot,self).__init__(site,prefix,showProgress=showProgress,debug=debug) 
        pass
    
    def submit(self, parameters):
        ''' submit the request with the given parameters
        Args:
            parameters(list): the parameters to use for the SMW API request
        Returns:
            dict: the submit result
        '''
        if not "Request" in sys.modules:
            from pywikibot.data.api import Request
                
        request=Request(site=self.site,parameters=parameters)
        result=None
        try: 
            result=request.submit()    
        except pywikibot.exceptions.TimeoutError as _toe:
            msg=f"submit for {self.site} failed due to pywikibot TimeoutError"
            raise Exception(msg)
            pass
        return result
    
    def info(self):
        """ see https://www.semantic-mediawiki.org/wiki/Help:API:smwinfo"""
        parameters={"action": "smwinfo"}
        return self.submit(parameters)
    
    def rawquery(self,ask:str,limit=None):
        '''
         send a query see https://www.semantic-mediawiki.org/wiki/Help:Inline_queries#Parser_function_.23ask
        
        Args:
            ask(str): the SMW ASK query as it would be used in MediaWiki markup
        '''
        # allow usage of original Wiki ask content - strip all non needed parts
        fixedAsk=self.fixAsk(ask)
        # set parameters for request
        parameters={"action": "ask","query":fixedAsk}
        result=self.submit(parameters)
        return result
    
    def query(self,ask,limit=None):
        '''
        send a query and deserialize it
        '''
        rawresult=self.rawquery(ask,limit=limit)
        result=self.deserialize(rawresult)
        return result


class QueryResultSizeExceedException(BaseException):
    """Raised if the results of a query can not completely be queried due to SMW result limits."""
    def __init__(self, result=[], message="Query can not completely be queried due to SMW result limits."):
        super().__init__(message)
        self.result=result

    def getResults(self):
        """ Returns the queried results before the exception was raised """
        return self.result





from typing import List, Union
from lodstorage.jsonable import JSONAble
from wikibot.wikiclient import WikiClient


class PageRevision(JSONAble):
    """
    Represents the metadata of a mediawiki page revision
    """

    @classmethod
    def getSamples(cls):
        samples=[
            {'revid': 7056,
             'parentid': 0,
             'user': '127.0.0.1',
             'anon': '',
             'userid': 0,
             'timestamp': '2008-10-14T21:23:09Z',
             'size': 6905,
             'comment': 'Event created',
             },
            {
                "revid": 8195,
                "parentid": 8194,
                "user": "Wf",
                "timestamp": "2021-11-11T12:50:31Z",
                "size": 910, "comment": ""
            }]
        return samples

    def __repr__(self):
        props = ', '.join([f"{k}={str(v)}" for k,v in self.__dict__.items()])
        return f"{self.__class__.__name__}({props})"


class PageHistory:
    """
    Represents the history of a page
    """

    def __init__(self, pageTitle:str, wikiId:str, debug:bool=False):
        """
        Constructor

        Args:
            pageTitle(str): name of the page
            wikiId(str): id of the wiki the page is located
            debug(bool): If True show debug messages
        """
        self.debug = debug
        self.pageTitle = pageTitle
        self.wikiClient = WikiClient.ofWikiId(wikiId, debug=self.debug)
        self.page = self.wikiClient.getPage(pageTitle)
        self.revisions = self._getRevisions()

    def _getRevisions(self) -> List[PageRevision]:
        """
        Get the revisions of the page as PageRevision object

        Returns:
            List of PageRevisions of the page
        """
        revisions = []
        for revisionRecord in self.page.revisions(prop="ids|timestamp|user|userid|comment|size"):
            revision = PageRevision()
            revision.fromDict(revisionRecord)
            revisions.append(revision)
        return revisions

    def exists(self) -> bool:
        """
        Checks if the page exists
        Assumption: If the page exists than ot exists at least one revision entry

        Returns:
            True if the page exists otherwise False
        """
        return len(self.revisions) > 0

    def getFirstUser(self, reverse:bool=False, limitedUserGroup:List[str]=None) -> Union[str, None]:
        """
        Returns the first user in the revisions

        Args:
            reverse(bool): If False start the search at the oldest entry. Otherwise, search from the newest to the oldest revision
            limitedUserGroup(list): limit the search to the given list. If None all users will be considered.

        Returns:
            str username that matches the search criterion
        """
        revisions = self.revisions
        revisions.sort(key=lambda r: int(getattr(r, "revid",0)))
        if reverse:
            revisions = reversed(revisions)
        for revision in revisions:
            user = getattr(revision, 'user', None)
            if user is None:
                continue
            if limitedUserGroup is None:
                return user
            elif user in limitedUserGroup:
                return user
        return None




'''
Created on 2020-11-02

@author: wf
'''
from wikibot.wikiuser import WikiUser
from mwclient import Site
from urllib.parse import urlparse
from wikibot.wiki import Wiki

class WikiClient(Wiki):
    '''
    access MediaWiki via mwclient library
    '''

    def __init__(self, wikiUser,debug=False):
        '''
        Constructor
        '''
        super(WikiClient,self).__init__(wikiUser,debug=debug) 
        self.wikiUser=wikiUser
        self.site=None
      
        
    def getSite(self):
        '''
        get my site
        '''
        if self.site is None:
            o=urlparse(self.wikiUser.url)
            scheme=o.scheme
            host=o.netloc
            path=o.path+self.wikiUser.scriptPath   
            path="%s/" % path
            self.site=Site(host=host, path=path, scheme=scheme)
        return self.site
        
    def login(self):
        '''
        login
        '''
        wu=self.wikiUser
        try:
            self.getSite().login(username=wu.user,password=wu.getPassword())
            return True
        except Exception as ex:
            if self.debug:
                print("login failed: %s" % str(ex))
            return False
        
    def getWikiMarkup(self,pageTitle):
        '''
        get the wiki markup code (text) for the given page Title
        
        Args:
            pageTitle(str): the title of the page to retrieve
            
        Returns:
            str: the wiki markup code for the page
        '''
        page=self.getPage(pageTitle)
        markup=page.text()
        return markup
    
    def getHtml(self,pageTitle):
        '''
        get the HTML code for the given page Title
        
        Args:
            pageTitle(str): the title of the page to retrieve
        '''
        api=self.getSite().api("parse",page=pageTitle)
        if not "parse" in api:
            raise Exception("could not retrieve html for page %s" % pageTitle)
        html=api["parse"]["text"]["*"]
        return html        
        
    def getPage(self,pageTitle):
        page=self.getSite().pages[pageTitle]
        return page
    
    def savePage(self,pageTitle,pageContent,pageSummary):
        '''
        save the page
        Args:
            pageTitle(str): the title of the page
            pageContent(str): the wikimarkup content
            pageSummary(str): 
        '''
        newPage=self.getPage(pageTitle)
        newPage.edit(pageContent,pageSummary)
        
    @staticmethod
    def getClients():
        clients={}
        for wikiUser in WikiUser.getWikiUsers().values():
            wikiClient=WikiClient(wikiUser)
            clients[wikiUser.wikiId]=wikiClient
        return clients
        
    @staticmethod
    def ofWikiId(wikiId,lenient=True,debug=False):
        wikiUser=WikiUser.ofWikiId(wikiId,lenient=lenient)
        wikibot=WikiClient(wikiUser,debug=debug)
        return wikibot
    
    @staticmethod
    def ofWikiUser(wikiUser):
        wikibot=WikiClient(wikiUser)
        return wikibot
        


'''
Created on 2020-11-12

@author: wf
'''
import sys
import wikibot.wikipush 
DEBUG=False

if __name__ == "__main__":
    if DEBUG:
        sys.argv.append("-d")
    sys.exit(wikibot.wikipush.mainUpload())


# ! important
# see https://stackoverflow.com/a/27868004/1497139
from setuptools import setup
from wikibot.version import Version
# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=Version.name,
    version=Version.version,

    packages=['wikibot',],
    classifiers=[
            'Programming Language :: Python',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            #'Programming Language :: Python :: 3.11'
    ],

    install_requires=[
      'gitpython',
      'jinja2',
      'pywikibot~=7.3',
      'pycryptodome~=3.15.0',
      'mwclient~=0.10.1',
      'mwparserfromhell~=0.6.4',
      'wikitextparser~=0.47.5'
    ],
    entry_points={
      'console_scripts': [
        'wikibackup = wikibot.wikipush:mainBackup',
        'wikiedit = wikibot.wikipush:mainEdit',
        'wikinuke = wikibot.wikipush:mainNuke',
        'wikipush = wikibot.wikipush:mainPush',
        'wikiquery = wikibot.wikipush:mainQuery',
        'wikiupload = wikibot.wikipush:mainUpload',
        'wikirestore = wikibot.wikipush:mainRestore',
        'wikiuser = wikibot.wikiuser:main',
      ],
    },
    author='Wolfgang Fahl',
    maintainer='Wolfgang Fahl',
    url='https://github.com/WolfgangFahl/py-3rdparty-mediawiki',
    license='Apache License',
    description='Wrapper for pywikibot with improvements for 3rd party wikis',
    long_description=long_description,
    long_description_content_type='text/markdown'
)



'''
Created on 2020-10-29
  @author:     wf
  @copyright:  Wolfgang Fahl. All rights reserved.

'''
from wikibot.version import Version
from wikibot.selector import Selector
from wikibot.wikiclient import WikiClient
from mwclient.image import Image
from wikibot.smw import SMWClient
#from difflib import Differ
import difflib
import datetime
from git import Repo
from pathlib import Path
import os
import re
import sys
from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
import json
from lodstorage.query import Query

class WikiPush(object):
    '''
    Push pages from one MediaWiki to another
    '''
    differ=None
    
    def __init__(self, fromWikiId, toWikiId=None,login=False,verbose=True,debug=False):
        '''
        Constructor
        '''
        self.verbose=verbose
        self.debug=debug
        self.fromWiki=None
        self.toWiki=None
        
        self.fromWikiId=fromWikiId
        if self.fromWikiId is not None:
            self.fromWiki=WikiClient.ofWikiId(fromWikiId,debug=self.debug)
        self.toWikiId=toWikiId
        if self.toWikiId is not None:
            self.toWiki=WikiClient.ofWikiId(toWikiId,debug=self.debug)
        if login and self.fromWikiId is not None:
            if not self.fromWiki.login():
                raise Exception("can't login to source Wiki %s" % fromWikiId )
        if self.toWiki is not None:
            if not self.toWiki.login():
                raise Exception("can't login to target Wiki %s" % toWikiId )
        
    def log(self,msg,end='\n'):
        if self.verbose:
            print (msg,end=end)
            
    def formatQueryResult(self,askQuery,wiki=None,limit=None,showProgress=False,queryDivision=1,outputFormat='lod', entityName="data",title:str=None):
        '''
        format the query result for the given askQuery.
        Args:
             askQuery(string): Semantic Media Wiki in line query https://www.semantic-mediawiki.org/wiki/Help:Inline_queries
            wiki(wikibot): the wiki to query - use fromWiki if not specified
            limit(int): the limit for the query (optional)
            showProgress(bool): true if progress of the query retrieval should be indicated (default: one dot per 50 records ...)
            queryDivision(int): Defines the number of subintervals the query is divided into (must be greater equal 1)
            outputFormat(str): output format of the query results - default format is lod
            entityName(str): the name of the entity
            title(str): the title of the query (if any)
        Returns:
            Query results in the requested outputFormat as string.
            If the requested outputFormat is not supported None is returned.
        '''
        pageRecords=self.queryPages(askQuery, wiki, limit, showProgress, queryDivision)
        outputFormat=outputFormat.lower()
        if outputFormat== "csv":
            return self.convertToCSV(pageRecords)
        elif outputFormat == "json":
            res = []
            for page in pageRecords.values():
                res.append(page)
            res_json = json.dumps({entityName: res}, default=str, indent=3)
            return res_json
        elif outputFormat == "lod":
            return [pageRecord for pageRecord in pageRecords.values()]
        else:
            if title is None:
                title=entityName
            query=Query(name=entityName,query=askQuery,title=title)
            qlod=[pageRecord for pageRecord in pageRecords.values()]
            doc=query.documentQueryResult(qlod, limit, tablefmt=outputFormat, withSourceCode=False)
            return doc
            #if self.debug:
            #    print(f"Format {outputFormat} is not supported.")

    def convertToCSV(self, pageRecords, separator=";"):
        """
        Converts the given pageRecords into a str in csv format
        ToDO: Currently does not support escaping of the separator and escaping of quotes
        Args:
            pageRecords: dict of dicts containing the printouts
            separator(char):
        Returns: str
        """
        res = ""
        printedHeaders = False
        for pageRecord in pageRecords.values():
            if not printedHeaders:
                for key in pageRecord.keys():
                    res = f"{res}{key}{separator}"
                res = f"{res[:-1]}\n"
                printedHeaders = True
            for printouts in pageRecord.values():
                res = f"{res}{printouts}{separator}"
            res = f"{res[:-1]}\n"   # remove last separator and end line
        return res

    def queryPages(self,askQuery,wiki=None,limit=None,showProgress=False, queryDivision=1) -> dict:
        '''
        query the given wiki for pagerecords matching the given askQuery
        
        Args:
            askQuery(string): Semantic Media Wiki in line query https://www.semantic-mediawiki.org/wiki/Help:Inline_queries
            wiki(wikibot): the wiki to query - use fromWiki if not specified
            limit(int): the limit for the query (optional)
            showProgress(bool): true if progress of the query retrieval should be indicated (default: one dot per 50 records ...)
            queryDivision(int): Defines the number of subintervals the query is divided into (must be greater equal 1)
        Returns:
            list: a list of pageRecords matching the given askQuery
        '''
        if wiki is None:
            wiki=self.fromWiki
        smwClient=SMWClient(wiki.getSite(),showProgress=showProgress, queryDivision=queryDivision,debug=self.debug)
        pageRecords=smwClient.query(askQuery,limit=limit)  
        return pageRecords
    
    def query(self,askQuery,wiki=None,queryField=None,limit=None,showProgress=False, queryDivision=1):
        '''
        query the given wiki for pages matching the given askQuery
        
        Args:
            askQuery(string): Semantic Media Wiki in line query https://www.semantic-mediawiki.org/wiki/Help:Inline_queries
            wiki(wikibot): the wiki to query - use fromWiki if not specified
            queryField(string): the field to select the pageTitle from
            limit(int): the limit for the query (optional)
            showProgress(bool): true if progress of the query retrieval should be indicated (default: one dot per 50 records ...)
        Returns:
            list: a list of pageTitles matching the given askQuery
        '''
        pageRecords=self.queryPages(askQuery, wiki, limit, showProgress, queryDivision)
        if queryField is None:
            return pageRecords.keys()
        # use a Dict to remove duplicates
        pagesDict={}
        for pageRecord in pageRecords.values():
            pagesDict[pageRecord[queryField]]=True
        return pagesDict.keys()
    
    def nuke(self,pageTitles,force=False):
        '''
        delete the pages with the given page Titles
        
        Args:
            pageTitles(list): a list of page titles to be transfered from the formWiki to the toWiki
            force(bool): True if pages should be actually deleted - dry run only listing pages is default
        '''
        total=len(pageTitles)
        self.log("deleting %d pages in %s (%s)" % (total,self.toWikiId,"forced" if force else "dry run"))
        for i,pageTitle in enumerate(pageTitles):
            try:
                self.log("%d/%d (%4.0f%%): deleting %s ..." % (i+1,total,(i+1)/total*100,pageTitle), end='')
                pageToBeDeleted=self.toWiki.getPage(pageTitle)   
                if not force:
                    self.log("👍" if pageToBeDeleted.exists else "👎")
                else:
                    pageToBeDeleted.delete("deleted by wiknuke")    
                    self.log("✅")
            except Exception as ex:
                msg=str(ex)
                self.log("❌:%s" % msg )
                
    @staticmethod
    def getDiff(text,newText,n=1,forHuman=True):
        #if WikiPush.differ is None:
        #    WikiPush.differ=Differ()
        # https://docs.python.org/3/library/difflib.html
        #  difflib.unified_diff(a, b, fromfile='', tofile='', fromfiledate='', tofiledate='', n=3, lineterm='\n')¶
        #diffs=WikiPush.differ.compare(,)
        textLines=text.split("\n")
        newTextLines=newText.split("\n")
        diffs=difflib.unified_diff(textLines,newTextLines,n=n)
        if forHuman:
            hdiffs=[]
            for line in diffs:
                unwantedItems=["@@","---","+++"]
                keep=True
                for unwanted in unwantedItems:
                    if unwanted in line:
                        keep=False
                if keep:
                    hdiffs.append(line)    
        else:
            hdiffs=diffs
        diffStr="\n".join(hdiffs)
        return diffStr
    
    @staticmethod
    def getModify(search:str,replace:str,debug:bool=False):
        '''
        get the modification function
        
        Args:
            search(str): the search string
            replace(str): the replace string
            debug(bool): if debug show
        '''
        if debug:
            print(f"search regex: {search}")
            print(f"replace regex: {replace}")
        searchRegex=r"%s" % search
        replaceRegex=r"%s" % replace
        modify=lambda text: re.sub(searchRegex,replaceRegex,text)
        return modify
                
    def edit(self,pageTitles,modify=None,context=1,force=False):
        '''
        edit the pages with the given page Titles
        
        Args:
            pageTitles(list): a list of page titles to be transfered from the formWiki to the toWiki
            force(bool): True if pages should be actually deleted - dry run only listing pages is default
        '''
        if modify is None:
            raise Exception("wikipush edit needs a modify function!")
        total=len(pageTitles)
        self.log("editing %d pages in %s (%s)" % (total,self.toWikiId,"forced" if force else "dry run"))
        for i,pageTitle in enumerate(pageTitles):
            try:
                self.log("%d/%d (%4.0f%%): editing %s ..." % (i+1,total,(i+1)/total*100,pageTitle), end='')
                pageToBeEdited=self.toWiki.getPage(pageTitle)   
                if not force and not pageToBeEdited.exists:
                    self.log("👎")
                else:
                    comment="edited by wikiedit" 
                    text=pageToBeEdited.text()
                    newText=modify(text)
                    if newText!=text:
                        if force:
                            pageToBeEdited.edit(newText,comment)  
                            self.log("✅")
                        else:
                            diffStr=self.getDiff(text, newText,n=context)
                            self.log(f"👍{diffStr}")
                    else:
                        self.log("↔")
            except Exception as ex:
                msg=str(ex)
                self.log(f"❌:{msg}")            
                
    def upload(self,files,force=False):
        '''
        push the given files
        Args:
            files(list): a list of filenames to be transfered to the toWiki
            force(bool): True if images should be overwritten if they exist
        '''
        total=len(files)
        self.log("uploading %d files to %s" % (total,self.toWikiId))
        for i,file in enumerate(files):
            try:
                self.log("%d/%d (%4.0f%%): uploading %s ..." % (i+1,total,(i+1)/total*100,file), end='')
                description="uploaded by wikiupload" 
                filename=os.path.basename(file)
                self.uploadImage(file, filename, description, force)
                self.log("✅")
            except Exception as ex:
                self.log("❌:%s" % str(ex) )
                
    def backup(self,pageTitles,backupPath=None,git=False,withImages=False):
        '''
        backup the given page titles
        Args:
            pageTitles(list): a list of page titles to be downloaded from the fromWiki
            git(bool): True if git should be used as a version control system
            withImages(bool): True if the image on a page should also be copied
        '''
        if backupPath is None:
            backupPath=self.getHomePath("wikibackup/%s" % self.fromWikiId)
        imageBackupPath="%s/images" % backupPath
        total=len(pageTitles)
        self.log("downloading %d pages from %s to %s" % (total,self.fromWikiId,backupPath))
        for i,pageTitle in enumerate(pageTitles):
            try:
                self.log("%d/%d (%4.0f%%): downloading %s ..." % (i+1,total,(i+1)/total*100,pageTitle), end='')
                page=self.fromWiki.getPage(pageTitle)
                wikiFilePath="%s/%s.wiki" % (backupPath,pageTitle)
                self.ensureParentDirectoryExists(wikiFilePath)
                with open (wikiFilePath,"w") as wikiFile:
                    wikiFile.write(page.text())
                self.log("✅")
                if isinstance(page,Image):
                    self.backupImages([page],imageBackupPath)
                if withImages:
                    self.backupImages(page.images(), imageBackupPath)    
                    
            except Exception as ex:
                self.log("❌:%s" % str(ex) )
        if git:
            gitPath="%s/.git" % backupPath
            if not os.path.isdir(gitPath):
                self.log("initializing git repository ...")
                repo=Repo.init(backupPath)
            else:
                repo=Repo(backupPath)
            self.log("committing to git repository")
            repo.git.add(all=True)
            timestamp=datetime.datetime.now().isoformat()
            repo.index.commit("auto commit by wikibackup at %s" % timestamp)
        
    def backupImages(self,imageList,imageBackupPath):
        '''
        '''
        for image in imageList:
            try:
                imagePath,filename=self.downloadImage(image,imageBackupPath)
            except Exception as ex:
                self.handleException(ex)
        
    def push(self,pageTitles,force=False,ignore=False,withImages=False):
        '''
        push the given page titles
        
        Args:
            pageTitles(list): a list of page titles to be transfered from the formWiki to the toWiki
            force(bool): True if pages should be overwritten if they exist
            ignore(bool): True if warning for images should be ignored (e.g if they exist)
            withImages(bool): True if the image on a page should also be copied
        '''
        total=len(pageTitles)
        self.log("copying %d pages from %s to %s" % (total,self.fromWikiId,self.toWikiId))
        for i,pageTitle in enumerate(pageTitles):
            try:
                self.log("%d/%d (%4.0f%%): copying %s ..." % (i+1,total,(i+1)/total*100,pageTitle), end='')
                page=self.fromWiki.getPage(pageTitle)
                if page.exists:
                    # is this an image?
                    if isinstance(page,Image):
                        self.pushImages([page],ignore=ignore)
                    else:
                        newPage=self.toWiki.getPage(pageTitle)
                        if not newPage.exists or force:
                            try:
                                comment="pushed from %s by wikipush" % self.fromWikiId
                                newPage.edit(page.text(),comment)
                                self.log("✅")
                                pageOk=True
                            except Exception as ex:
                                pageOk=self.handleException(ex, ignore)
                            if withImages and pageOk:
                                self.pushImages(page.images(),ignore=ignore)
                        else:
                            self.log("👎")
                else:
                    self.log("❌")
            except Exception as ex:
                self.log("❌:%s" % str(ex) )       
                
    def ensureParentDirectoryExists(self,filePath):
        # for pages that have a "/" in the name:
        directory = os.path.dirname(filePath)
        self.ensureDirectoryExists(directory)
        
    def ensureDirectoryExists(self,directory):
        '''
        make sure the given directory exists
        
        Args: 
            directory(str): the directory to check for existence
        '''
        Path(directory).mkdir(parents=True, exist_ok=True)
        
    def getHomePath(self,localPath):
        '''
        get the given home path
        '''
        homePath=f"{Path.home()}/{localPath}"
        self.ensureDirectoryExists(homePath)
        return homePath
                    
    def getDownloadPath(self):
        '''
        get the download path
        '''
        return self.getHomePath("Downloads/mediawiki")
        
            
    def pushImages(self,imageList,delim="",ignore=False):
        '''
        push the images in the given image List
        
        Args:
            imageList(list): a list of images to be pushed
            ignore(bool): True to upload despite any warnings.
        '''        
        for image in imageList:
            try:
                self.log("%scopying image %s ..." % (delim,image.name), end='')
                imagePath,filename=self.downloadImage(image)
                description=image.imageinfo['comment']
                try:
                    self.uploadImage(imagePath,filename,description,ignore)
                    self.log("✅")
                except Exception as ex:
                    self.handleAPIWarnings(ex.args[0],ignoreExists=ignore)
                if self.debug:
                    print(image.imageinfo)
            except Exception as ex:
                self.handleException(ex,ignore)
                
    def handleException(self,ex,ignoreExists=False):
        '''
        handle the given exception and ignore it if it includes "exists" and ignoreExists is True
        
        Args:
            ex(Exception): the exception to handle
            ignoreExists(bool): True if "exists" should be ignored
            
        Returns:
            bool: True if the exception was handled as ok False if it was logged as an error
        '''
        msg=str(ex)
        return self.handleWarning(msg,marker="❌",ignoreExists=ignoreExists)
        
    def handleAPIWarnings(self,warnings,ignoreExists=False):
        '''
        handle API Warnings
        
        Args:
            warnings(list): a list of API warnings
            ignoreExists(bool): ignore messages that warn about existing content

        Returns:
            bool: True if the exception was handled as ok False if it was logged as an error
        '''
        msg=""
        if warnings:
            if isinstance(warnings, str):
                msg = warnings
            else:
                for warning in warnings:
                    msg+="%s\n" % str(warning)
        return self.handleWarning(msg,ignoreExists=ignoreExists)
        
    def handleWarning(self,msg,marker="⚠️",ignoreExists=False):
        '''
        handle the given warning and ignore it if it includes "exists" and ignoreExists is True
        
        Args:
            msg(string): the warning to handle
            marker(string): the marker to use for the message
            ignoreExists(bool): True if "exists" should be ignored
            
        Returns:
            bool: True if the exception was handled as ok False if it was logged as an error
        '''
        #print ("handling warning %s with ignoreExists=%r" % (msg,ignoreExists))
        if ignoreExists and "exists" in msg:
            # shorten exact duplicate message
            if "exact duplicate" in msg:
                msg="exact duplicate"
            marker="👀"
        if not ignoreExists and "exists" in msg:
            msg="file exists (to overwrite existing files enable the ignore parameter)"
        self.log("%s:%s" % (marker,msg))
        return marker=="👀"
                
    def downloadImage(self,image,downloadPath=None):
        '''
        download the given image
        
        Args:
            image(image): the image to download
            downloadPath(str): the path to download to if None getDownloadPath will be used
        '''
        filename=image.name.replace("File:","")
        if downloadPath is None:
            downloadPath=self.getDownloadPath()
        imagePath="%s/%s" % (downloadPath,filename)
        self.ensureParentDirectoryExists(imagePath)
        with open(imagePath,'wb') as imageFile:
            image.download(imageFile)
        return imagePath,filename
                            
    def uploadImage(self,imagePath,filename,description,ignoreExists=False):
        '''
        upload an image
        
        Args:
            imagePath(str): the path to the image
            filename(str): the filename to use
            description(str): the description to use
            ignoreExists(bool): True if it should be ignored if the image exists
        '''
        with open(imagePath,'rb') as imageFile:
            warnings=None
            response=self.toWiki.site.upload(imageFile,filename,description,ignoreExists)
            if 'warnings' in response:
                warnings=response['warnings']
            if 'upload' in response and 'warnings' in response['upload']:
                warningsDict=response['upload']['warnings']
                warnings=[]
                for item in warningsDict.items():
                    warnings.append(str(item))
            if warnings:
                raise Exception(warnings)

    def restore(self, pageTitles=None, backupPath=None, listFile=None, stdIn=False):
        """
        restore given page titles from local backup
        If no page titles are given the whole backup is restored.
        Args:
            pageTitles(list): a list of pageTitles to be restored to toWiki. If None -> full restore of backup
            backupPath(str): path to backup location
        """
        if stdIn:
            backupPath = os.path.dirname(pageTitles[0].strip())
            pageTitlesfix= []
            for i in pageTitles:
                pageTitlesfix.append(os.path.basename(i.strip().replace('.wiki','')))
            pageTitles = pageTitlesfix
        elif listFile is not None:
            f = open(listFile, 'r')
            allx = f.readlines()
            pageTitles = []
            for i in allx:
                pageTitles.append(os.path.basename(i.strip()).replace('.wiki',''))
        else:
            if backupPath is None:
                backupPath=self.getHomePath("wikibackup/%s" % self.toWikiId)
            if pageTitles is None:
                pageTitles = []
                for path, subdirs, files in os.walk(backupPath):
                    for name in files:
                        filename = os.path.join(path, name)[len(backupPath)+1:]
                        if filename.endswith(".wiki"):
                            pageTitles.append(filename[:-len(".wiki")])
        total = len(pageTitles)
        self.log("restoring %d pages from %s to %s" % (total, backupPath, self.toWikiId))
        for i,pageTitle in enumerate(pageTitles):
            try:
                self.log("%d/%d (%4.0f%%): restore %s ..." % (i + 1, total, (i + 1) / total * 100, pageTitle),end='')
                wikiFilePath = "%s/%s.wiki" % (backupPath, pageTitle)
                with open(wikiFilePath, mode='r') as wikiFile:
                    page_content = wikiFile.read()
                    page = self.toWiki.getPage(pageTitle)
                    page.edit(page_content, f"modified through wikirestore by {self.toWiki.wikiUser.user}")
                self.log("✅")
            except Exception as ex:
                self.log("❌:%s" % str(ex) )

__version__ = Version.version
__date__ = Version.date
__updated__ = Version.updated
DEBUG=False

def mainNuke(argv=None):
    main(argv,mode='wikinuke')

def mainEdit(argv=None):
    main(argv,mode='wikiedit')

def mainPush(argv=None):
    main(argv,mode='wikipush')

def mainQuery(argv=None):
    main(argv,mode='wikiquery')

def mainUpload(argv=None):
    main(argv,mode='wikiupload')

def mainBackup(argv=None):
    main(argv,mode='wikibackup')

def mainRestore(argv=None):
    main(argv,mode='wikirestore')

def main(argv=None,mode='wikipush'): # IGNORE:C0111
    '''main program.'''

    if argv is None:
        argv = sys.argv[1:]

    program_name = mode
    program_version = "v%s" % __version__
    program_build_date = str(__updated__)
    program_version_message = '%%(prog)s %s (%s)' % (program_version, program_build_date)
    program_shortdesc = "wikipush"
    user_name="Wolfgang Fahl"

    program_license = '''%s

  Created by %s on %s.
  Copyright 2020 Wolfgang Fahl. All rights reserved.

  Licensed under the Apache License 2.0
  http://www.apache.org/licenses/LICENSE-2.0

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

''' % (program_shortdesc,user_name, str(__date__))

    try:
        # Setup argument parser
        parser = ArgumentParser(description=program_license, formatter_class=RawDescriptionHelpFormatter)
        parser.add_argument("-d", "--debug", dest="debug",   action="store_true", help="set debug level [default: %(default)s]")
        parser.add_argument('-V', '--version', action='version', version=program_version_message)
        if mode=="wikipush":
            parser.add_argument("-l", "--login", dest="login", action='store_true', help="login to source wiki for access permission")
            parser.add_argument("-s", "--source", dest="source", help="source wiki id", required=True)
            parser.add_argument("-f", "--force", dest="force", action='store_true', help="force to overwrite existing pages")
            parser.add_argument("-i", "--ignore", dest="ignore", action='store_true', help="ignore upload warnings e.g. duplicate images")
            parser.add_argument("-wi", "--withImages", dest="withImages", action='store_true', help="copy images on the given pages")
        elif mode=="wikibackup":
            parser.add_argument("-g", "--git", dest="git", action='store_true', help="use git for version control")
            parser.add_argument("-l", "--login", dest="login", action='store_true', help="login to source wiki for access permission")
            parser.add_argument("-s", "--source", dest="source", help="source wiki id", required=True)
            parser.add_argument("-wi", "--withImages", dest="withImages", action='store_true', help="copy images on the given pages")
            parser.add_argument("--backupPath", dest="backupPath", help="path where the backup should be stored", required=False)
        elif mode=="wikinuke":
            parser.add_argument("-f", "--force", dest="force", action='store_true', help="force to delete pages - default is 'dry' run only listing pages")
        elif mode=="wikiedit":
            parser.add_argument("--search", dest="search", help="search pattern", required=True)
            parser.add_argument("--replace", dest="replace", help="replace pattern", required=True)
            parser.add_argument("--context", dest="context",type=int, help="number of context lines to show in dry run diff display",default=1)
            parser.add_argument("-f", "--force", dest="force", action='store_true', help="force to edit pages - default is 'dry' run only listing pages")
        elif mode=="wikiquery":
            parser.add_argument("-l", "--login", dest="login", action='store_true', help="login to source wiki for access permission")
            parser.add_argument("-s", "--source", dest="source", help="source wiki id", required=True)
            parser.add_argument("--format", dest="format", default='json', help="format to use for query result csv,json,lod or any of the tablefmt options of https://pypi.org/project/tabulate/")
            parser.add_argument("--entityName", dest="entityName", default='data', help="name of the entites that are queried - only needed for some output formats - default is 'data'")
        elif mode=="wikiupload":
            parser.add_argument("--files", nargs='+', help="list of files to be uploaded", required=True)
            parser.add_argument("-f", "--force", dest="force", action='store_true', help="force to (re)upload existing files - default is false")
            pass
        elif mode=="wikirestore":
            parser.add_argument("--listFile", dest="listFile", help="List of pages to restore", required=False)
            parser.add_argument("--backupPath", dest="backupPath", help="path the backup is stored", required=False)
            parser.add_argument("-s", "--source", dest="source", help="source wiki id", required=False)
            parser.add_argument("-l", "--login", dest="login", action='store_true', help="login to source wiki for access permission")
            parser.add_argument('-stdinp', dest="stdinp", action='store_true',help='Use the input from STD IN using pipes')
        if mode in  ["wikipush","wikiedit","wikinuke","wikibackup","wikiquery","wikirestore"]:
            parser.add_argument("--limit",dest="limit",type=int,help="limit for query")
            parser.add_argument("--progress",dest="showProgress",action='store_true',help="shows progress for query")
            parser.add_argument("-q", "--query", dest="query", help="select pages with given SMW ask query", required=False)
            parser.add_argument("--queryFile", dest="queryFile", help="file the query should be read from")
            parser.add_argument("-qf", "--queryField",dest="queryField",help="query result field which contains page")
            parser.add_argument("-p", "--pages", nargs='+', help="list of page Titles to be pushed", required=False)
            parser.add_argument("-ui", "--withGUI", dest="ui", help="Pop up GUI for selection", action="store_true",required=False)
            parser.add_argument("-qd", "--queryDivision", default=1, dest="queryDivision", type=int, help="divide query into equidistant subintervals to limit the result size of the individual queries", required=False)
        if mode in  ["wikiquery"]:
            parser.add_argument("--title",help="the title for the query")
        if not mode in ["wikibackup", "wikiquery"]:
            parser.add_argument("-t", "--target", dest="target", help="target wiki id", required=True)
        # Process arguments
        args = parser.parse_args(argv)
        if hasattr(args,"queryDivision"):
            if args.queryDivision < 1:
                raise ValueError("queryDivision argument must be greater equal 1")

        if mode=="wikipush":
            wikipush=WikiPush(args.source,args.target,login=args.login,debug=args.debug)
            queryWiki=wikipush.fromWiki
        elif mode=="wikibackup":
            wikipush=WikiPush(args.source,None,login=args.login,debug=args.debug)
            queryWiki=wikipush.fromWiki
        elif mode=="wikiquery":
            wikipush=WikiPush(args.source,None,login=args.login,debug=args.debug)
            queryWiki=wikipush.fromWiki
        elif mode=="wikiupload":
            wikipush=WikiPush(None,args.target,debug=args.debug)
        elif mode=="wikirestore":
            wikipush=WikiPush(args.source,args.target,login=args.login,debug=args.debug)
            queryWiki=wikipush.fromWiki
        else:
            wikipush=WikiPush(None,args.target,debug=args.debug)
            queryWiki=wikipush.toWiki
        if mode=="wikiupload":
            wikipush.upload(args.files,args.force)
        else:
            pages=None
            if args.pages:
                pages = args.pages
            elif hasattr(args,"stdinp"):
                if args.stdinp:
                    pages = sys.stdin.readlines()
            elif args.query or args.queryFile:
                if args.query:
                    query = args.query
                else:
                    with open(args.queryFile, 'r') as queryFile:
                        query=queryFile.read()
                if mode=="wikiquery":
                    formatedQueryResults = wikipush.formatQueryResult(query,wiki=queryWiki,limit=args.limit,showProgress=args.showProgress, queryDivision=args.queryDivision,outputFormat=args.format, entityName=args.entityName,title=args.title)
                    if formatedQueryResults:
                        print(formatedQueryResults)
                    else:
                        print(f"Format {args.format} is not supported.")
                else:
                    pages=wikipush.query(query,wiki=queryWiki,queryField=args.queryField,limit=args.limit,showProgress=args.showProgress, queryDivision=args.queryDivision)
            if pages is None:
                if mode=="wikiquery":
                    # we are finished
                    pass
                elif mode=="wikirestore":
                    if args.pages is None and args.queryFile is None and args.query is None:
                        wikipush.restore(pageTitles=None,backupPath=args.backupPath,listFile=args.listFile)
                else:
                    raise Exception("no pages specified - you might want to use the -p, -q or --queryFile option")
            else:
                if args.ui and len(pages) > 0:
                    pages = Selector.select(pages, action=mode.lower().lstrip("wiki")[0].upper() + mode.lstrip("wiki")[1:],
                                            description='GUI program for the mode ' + mode,title=mode)
                    if pages == 'Q': #If GUI window is closed, end the program
                        sys.exit(0)
                if mode=="wikipush":
                    wikipush.push(pages,force=args.force,ignore=args.ignore,withImages=args.withImages)
                elif mode=="wikibackup":
                    wikipush.backup(pages,git=args.git,withImages=args.withImages,backupPath=args.backupPath)
                elif mode=='wikinuke':
                    wikipush.nuke(pages,force=args.force)
                elif mode=='wikiedit':
                    modify=WikiPush.getModify(args.search,args.replace,args.debug)
                    wikipush.edit(pages,modify=modify,context=args.context,force=args.force)
                elif mode=="wikirestore":
                    wikipush.restore(pages,backupPath=args.backupPath,stdIn=args.stdinp)
                else:
                    raise Exception("undefined wikipush mode %s" % mode)
        
    except KeyboardInterrupt:
        ### handle keyboard interrupt ###
        return 1
    except Exception as e:
        if DEBUG:
            raise(e)
        indent = len(program_name) * " "
        sys.stderr.write(program_name + ": " + repr(e) + "\n")
        sys.stderr.write(indent + "  for help use --help")
        return 2
                
if __name__ == "__main__":
    if DEBUG:
        sys.argv.append("-d")
    sys.exit(mainPush())
